@@ -87,11 +87,8 @@ select_track (void)
     return FALSE;
   }
 
-  gst_element_set_state (pipeline, GST_STATE_PAUSED);
   cd = gst_bin_get_by_name_recurse_up (GST_BIN (pipeline), "cd-source");
   gst_element_seek (pipeline, 1.0, gst_format_get_by_nick ("track"), GST_SEEK_FLAG_FLUSH, GST_SEEK_TYPE_SET, seek_to_track, GST_SEEK_TYPE_NONE, -1);
-  /* get_state neccessary because of bug #326311 */
-  gst_element_get_state (pipeline, NULL, NULL, -1);
   current_track = seek_to_track;
   seek_to_track = -1;
 
@@ -263,6 +260,8 @@ cb_change_button (gpointer data)
 {
   button_change_id = 0;
 
+  /* TODO: find out why GTK+ needs this to work (see #364371) */
+  gtk_button_set_label (GTK_BUTTON (data), _("Play"));
   gtk_button_set_label (GTK_BUTTON (data), GTK_STOCK_MEDIA_PLAY);
 
   /* once */
@@ -305,6 +304,8 @@ cb_state (GstBus *bus, GstMessage *message, gpointer user_data)
     gtk_statusbar_push (GTK_STATUSBAR (statusbar), 0, "");
     current_track = -1;
   } else if (transition == GST_STATE_CHANGE_PAUSED_TO_PLAYING) {
+    /* TODO: find out why GTK+ needs this to work (see #364371) */
+    gtk_button_set_label (GTK_BUTTON (play_button), _("Pause"));
     gtk_button_set_label (GTK_BUTTON (play_button), GTK_STOCK_MEDIA_PAUSE);
     if (id)
       g_source_remove (id);
@@ -357,7 +358,7 @@ setup (GError **err)
     g_signal_connect (bus, "message::error", G_CALLBACK (cb_error), NULL);
     g_signal_connect (bus, "message::state-changed", G_CALLBACK (cb_state), NULL);
 
-    cdp = gst_element_factory_make (CD_SRC, "cd-source");
+    cdp = gst_element_make_from_uri (GST_URI_SRC, "cdda://1", "cd-source");
     if (!cdp) {
       gst_object_unref (GST_OBJECT (pipeline));
       pipeline = NULL;
@@ -367,8 +368,12 @@ setup (GError **err)
     /* do not set to 1, that messes up buffering. 2 is enough to keep
      * noise from the drive down. */
     /* TODO: will not notice drive changes, should monitor */
+    if (g_object_class_find_property (G_OBJECT_GET_CLASS (cdp), "read-speed") != NULL) {
+    	g_object_set (G_OBJECT (cdp),
+    	              "read-speed", 2,
+                      NULL);
+    }
     g_object_set (G_OBJECT (cdp),
-                  "read-speed", 2,
                   "device", nautilus_burn_drive_get_device (drive),
                   NULL);
 
@@ -638,11 +643,7 @@ on_seek_release (GtkWidget * scale, GdkEventButton * event, gpointer user_data)
   cd = gst_bin_get_by_name_recurse_up (GST_BIN (pipeline), "cd-source");
   seeking = FALSE;
 
-  /* set_state/get_state neccessary because of bug #326311 */
-  gst_element_set_state (pipeline, GST_STATE_PAUSED);
   gst_element_seek (pipeline, 1.0, GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH, GST_SEEK_TYPE_SET, slen * val, GST_SEEK_TYPE_NONE, -1);
-  gst_element_get_state (pipeline, NULL, NULL, -1);
-  gst_element_set_state (pipeline, GST_STATE_PLAYING);
 
   return FALSE;
 }
